use anyhow::Result;
use git2::{Repository, Signature};
use std::path::Path;

pub const TAGS: &'static [&'static str] = &["refs/tags/*:refs/tags/*"];

pub fn pull(repo: &Repository, token: String, what: &[&str]) -> Result<()> {
    let mut remote = repo.find_remote("origin")?;
    let mut fo = git2::FetchOptions::new();

    let mut callbacks = git2::RemoteCallbacks::new();
    callbacks.credentials(|_u, _username, allowed_types| {
        if allowed_types.contains(git2::CredentialType::SSH_KEY) {
            let user = "git";
            git2::Cred::ssh_key_from_agent(user)
        } else {
            git2::Cred::userpass_plaintext(&token, "")
        }
    });

    fo.remote_callbacks(callbacks);

    remote.connect(git2::Direction::Fetch)?;
    remote.fetch(what, Some(&mut fo), None)?;
    Ok(())
}

pub fn push(repo: &Repository, token: String, what: &[&str]) -> Result<()> {
    let mut remote = repo.find_remote("origin")?;
    let mut po = git2::PushOptions::new();

    let mut callbacks = git2::RemoteCallbacks::new();
    callbacks.credentials(|_u, _username, allowed_types| {
        if allowed_types.contains(git2::CredentialType::SSH_KEY) {
            let user = "git";
            git2::Cred::ssh_key_from_agent(user)
        } else {
            git2::Cred::userpass_plaintext(&token, "")
        }
    });

    po.remote_callbacks(callbacks);

    remote.connect(git2::Direction::Push)?;
    remote.push(what, Some(&mut po))?;
    Ok(())
}

pub fn commit(
    repo: &Repository,
    sig: &git2::Signature,
    msg: &str,
    files: &[&str],
) -> Result<()> {
    let mut index = repo.index()?;
    for file in files {
        index.add_path(Path::new(file))?;
    }

    let oid = index.write_tree()?;
    let tree = repo.find_tree(oid)?;
    repo.commit(Some("HEAD"), &sig, &sig, &msg, &tree, &[])?;

    Ok(())
}

pub fn tag_version(repo: &Repository, tag: String, desc: String, sig: &git2::Signature) -> Result<()> {
    let obj = repo.revparse_single("HEAD")?;
    repo.tag(&tag, &obj, &sig, &desc, false)?;

    Ok(())
}

pub fn has_tag(repo: &Repository, tag: String) -> Result<bool> {
    let tags = repo.tag_names(Some(&tag))?;

    for tag_obj in tags.iter() {
        if tag_obj.is_none() {
            continue;
        }

        let tag_name = tag_obj.unwrap();
        if tag == tag_name.to_string() {
            return Ok(true);
        }
    }

    Ok(false)
}

#[cfg(test)]
mod tests {
    use anyhow::Result;
    use git2::*;
    use std::{fs::File, io::Write};
    use tempfile::tempdir;

    #[test]
    fn has_tag() -> Result<()> {
        const TAG: &'static str = "0.1.0";
        let dir = tempdir()?;
        let repo = Repository::init(&dir)?;
        let mut fout = File::create(&dir.path().join("VERSION"))?;
        write!(fout, "{}", TAG)?;
        drop(fout);

        let sig = &Signature::now("Gitea Release Tool", "gitea-release@tulpa.dev")?;
        super::commit(&repo, &sig, "test commit please ignore", &["VERSION"])?;

        super::tag_version(&repo, TAG.into(), format!("version {}", TAG), &sig)?;
        assert!(super::has_tag(&repo, TAG.into())?);

        Ok(())
    }
}
